/* 
   COPYRIGHT 2008 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - ssl_utils.c */
/* rdk@krupczak.org */

/* utils for supporting XMP over ssl */

#ifdef WIN32
#include <windows.h>
#include <Iphlpapi.h>
#include <direct.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include <openssl/ssl.h>
#include <openssl/rsa.h>
#include <openssl/engine.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>

#include "xmp.h"
#include "xmp_session.h"

SSL_CTX *ctx = NULL;   /* SSL context */

/* ************************************************************* */
/* not perfect algorithm for coming up with a FQDN but good enough
   for doing our certs */

#ifdef WIN32

int getdomainname(char *domainname, int maxlen)
{
     FIXED_INFO fi;
     ULONG fisize;

     if (domainname != NULL)
        domainname[0] = 0;

     /* use IP Helper functions new in NT4/SP4+ */
     /* need to install platform SDK to get the include file(s) */
     fisize = sizeof(fi);
     if (GetNetworkParams(&fi,&fisize) < 0) {
        fprintf(stderr,"getdomainname: failed to get network parameters\n");
        return -1;
     }

     strncpy(domainname,fi.DomainName,maxlen-1);
     domainname[maxlen-1] = 0;

     return 1;
}

#endif /* WIN32 */

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */

struct entry
{
  char *key;
  char *value;
};

#define ENTRY_COUNT 6

struct entry entries[ENTRY_COUNT] = {
  {"countryName", "US"},
  {"stateOrProvinceName", "Georgia"},
  {"localityName", "Smyrna"},
  {"organizationName", "Krupczak.org, LLC"},
  {"organizationalUnitName", "Cartographer"},
#define ENTRY_FQDN  5 
  {"commonName", "FQDN"},
};

int create_cert_request(char *fqdn, char *certfile, char *csrfile)
{
     int i;
     X509_REQ *req;
     X509_NAME *subj;
     EVP_PKEY *pkey;
     EVP_MD *digest;
     FILE *fp;

     if ((fqdn == NULL) || (fqdn[0] == 0)) {
        fprintf(stderr,"create_cert_request: no FQDN\n");
        return -1;
     }
     if ((csrfile == NULL) || (csrfile[0] == 0)) {
        fprintf(stderr,"create_cert_request: no CSR file\n");
        return -1;
     }

     /* first read in the private key from the eventual certfile */
     if (!(fp = fopen(certfile, "r"))) {
        fprintf(stderr,"Error reading private key file\n");
        return -1;
     }

     /* read private key passing password if its needed; we dont
        password encrypt our files (yet) */
     if (!(pkey = PEM_read_PrivateKey(fp, NULL, NULL, "xxx"))) {
        fprintf(stderr,"Error reading private key in file\n");
        fclose (fp);
        return -1;
     }
     fclose (fp);

     /* set the FQDN as the value for entry[X] 
      */
     entries[ENTRY_FQDN].value = fqdn;      

     /* create a new request and add the key to it */
     if (!(req = X509_REQ_new())) {
        fprintf(stderr,"Failed to create X509_REQ object\n");
        return -1;
     }

     X509_REQ_set_pubkey(req, pkey);

     /* assign the subject name */
     if (!(subj = X509_NAME_new())) {
        fprintf(stderr,"Failed to create X509_NAME object\n");
        return -1;
     }

     for (i = 0; i < ENTRY_COUNT; i++) {
         int nid;
         X509_NAME_ENTRY *ent;

         if ((nid = OBJ_txt2nid(entries[i].key)) == NID_undef)	{
	    fprintf (stderr, "Error finding NID for %s\n", entries[i].key);
            return -1;
	 }
         if (!(ent = X509_NAME_ENTRY_create_by_NID(NULL, nid, MBSTRING_ASC,
	                                            entries[i].value, -1))) {
	    fprintf(stderr,"Error creating Name entry from NID\n");
            return -1;
         }
         if (X509_NAME_add_entry(subj, ent, -1, 0) != 1) {
	    fprintf(stderr,"Error adding entry to Name\n");
            return -1;
         }
     }

     if (X509_REQ_set_subject_name(req, subj) != 1) {
        fprintf(stderr,"Error adding subject to request");
        return -1;
     }

     /* add an extension for the FQDN we wish to have */
     {
     X509_EXTENSION *ext;
     STACK_OF (X509_EXTENSION) * extlist;
     char *name = "subjectAltName";
     char value[512];

     sprintf(value,"DNS:%s",fqdn);

     extlist = sk_X509_EXTENSION_new_null ();

     if (!(ext = X509V3_EXT_conf(NULL, NULL, name, value))) {
        fprintf(stderr,"Error creating subjectAltName extension\n");
     }

     sk_X509_EXTENSION_push(extlist, ext);

     if (!X509_REQ_add_extensions (req, extlist)) {
        fprintf(stderr,"Error adding subjectAltName to the request\n");
        return -1;
     }
     sk_X509_EXTENSION_pop_free(extlist, X509_EXTENSION_free);
     }

     /* pick the correct digest and sign the request;
        we know we are using RSA for our private keys
        so set digest accordingly */

     digest = (EVP_MD *)EVP_sha1();

     if (!(X509_REQ_sign(req, pkey, digest))) {
        fprintf(stderr,"Error signing request");
        return -1;
     }

     /* write the completed request */
     if (!(fp = fopen(csrfile, "w"))) {
        fprintf(stderr,"Error writing to request file\n");
        return -1;
     }
     if (PEM_write_X509_REQ(fp, req) != 1) {
        fprintf(stderr,"Error while writing request\n");
        fclose (fp);
        return -1;
     }
     fclose (fp);

     EVP_PKEY_free(pkey);
     X509_REQ_free(req);

     return 1;
}

/* ************************************************************* */
/* generate cert given our Cartographer CA and CSR */

#define EXT_COUNT 5

struct entry ext_ent[EXT_COUNT] = {
  {"basicConstraints", "CA:FALSE"},
  {"nsComment", "\"Cartographer Generated Certificate\""},
  {"subjectKeyIdentifier", "hash"},
  {"authorityKeyIdentifier", "keyid,issuer:always"},
  {"keyUsage", "nonRepudiation,digitalSignature,keyEncipherment"}
};

#define CERT_DAYS    3650
#define CERT_SECS    (60*60*24*CERT_DAYS)

int generate_cert(char *certfile, char *csrfile)
{
     int i, subjAltName_pos;
     long serial = 1;
     EVP_PKEY *pkey, *CApkey;
     const EVP_MD *digest;
     X509 *cert, *CAcert;
     X509_REQ *req;
     X509_NAME *name;
     X509V3_CTX ctx;
     X509_EXTENSION *subjAltName;
     STACK_OF(X509_EXTENSION) *req_exts;
     FILE *fp;
     char namebuf[4096];

     /* read in the request */
     if (!(fp = fopen(csrfile, "r"))) {
        fprintf(stderr,"Error reading request file\n");
     }
     if (!(req = PEM_read_X509_REQ(fp, NULL, NULL, NULL))) {
        fprintf(stderr,"Error reading request in file\n");
        fclose (fp);
        return -1;
     }
     fclose (fp);

     /* verify signature on the request */
     if (!(pkey = X509_REQ_get_pubkey(req))) {
        fprintf(stderr,"Error getting public key from request\n");
        return -1;
     }
     if (X509_REQ_verify(req, pkey) != 1) {
        fprintf(stderr,"Error verifying signature on certificate\n");
        return -1;
     }

     /* read in the CA certificate */
     if (!(fp = fopen("cartographer.pem", "r"))) {
        fprintf(stderr,"Error reading CA certificate file\n");
        return -1;
     }
     if (!(CAcert = PEM_read_X509(fp, NULL, NULL, NULL))) {
        fprintf(stderr,"Error reading CA certificate in file\n");
        fclose(fp);
        return -1;
     }
     fclose(fp);

     /* read in the CA private key */
     if (!(fp = fopen("cartographer.pem","r"))) {
        fprintf(stderr,"Error reading CA private key file\n");
        return -1;
     }
     /* read encrypted private key */
     if (!(CApkey = PEM_read_PrivateKey(fp, NULL, NULL, "housekeeping finishing, unloading plugins"))) {
        fprintf(stderr,"Error reading CA private key in file\n");
        fclose (fp);
        return -1;
     }
     fclose (fp);

     /* print out and log the subject name and subject alt name extension */
     if (!(name = X509_REQ_get_subject_name(req))) {
        fprintf(stderr,"Error getting subject name from request\n");
        return -1;
     }
     namebuf[0] = 0;
     X509_NAME_oneline(name,namebuf,sizeof(namebuf)-1);

     /* fprintf(stderr,"Name: %s\n",namebuf); */

     if (!(req_exts = X509_REQ_get_extensions(req))) {
        fprintf(stderr,"Error getting the request's extensions\n");
        return -1;
     }
     subjAltName_pos = X509v3_get_ext_by_NID(req_exts,
					   OBJ_sn2nid ("subjectAltName"), -1);
     subjAltName = X509v3_get_ext(req_exts, subjAltName_pos);

#if 0
     {
     BIO *bufout;
     char *bufoutptr;

     bufout = BIO_new(BIO_s_mem());
     ret = X509V3_EXT_print(bufout,subjAltName,0,0);
     BIO_flush(bufout);
     bufoutptr = 0;
     BIO_get_mem_ptr(bufout,&bufoutptr);
     fprintf(stderr,"subjAltName: %d %lx %s\n",ret,bufoutptr,bufoutptr);
     for (i=0; i<20; i++) {
         printf("'%c' ",bufoutptr[i]);
     }
     printf("\n");
     BIO_free(bufout);

     bufout = BIO_new_fp(stdout,BIO_NOCLOSE);
     X509V3_EXT_print(bufout,subjAltName,0,0);
     fputc('\n',stdout);
     }
#endif

     /* create new certificate */
     if (!(cert = X509_new())) {
        fprintf(stderr,"Error creating X509 object\n");
        return -1;
     }

     /* set version number for the certificate (X509v3) and the serial number */
     if (X509_set_version(cert, 2L) != 1) {
        fprintf(stderr,"Error settin certificate version\n");
        return -1;
     }
     ASN1_INTEGER_set (X509_get_serialNumber (cert), serial++);

     /* set issuer and subject name of the cert from the req and the CA */
     if (!(name = X509_REQ_get_subject_name(req))) {
        fprintf(stderr,"Error getting subject name from request\n");
        return -1;
     }
     if (X509_set_subject_name(cert, name) != 1) {
        fprintf(stderr,"Error setting subject name of certificate\n");
        return -1;
     }
     if (!(name = X509_get_subject_name(CAcert))) {
        fprintf(stderr,"Error getting subject name from CA certificate\n");
        return -1;
     }
     if (X509_set_issuer_name(cert, name) != 1) {
        fprintf(stderr,"Error setting issuer name of certificate\n");
        return -1;
     }

     /* set public key in the certificate */
     if (X509_set_pubkey(cert, pkey) != 1) {
        fprintf(stderr,"Error setting public key of the certificate\n");
        return -1;
     }

     /* set duration for the certificate in terms of secs until expiration */
     if (!(X509_gmtime_adj(X509_get_notBefore(cert), 0))) {
        fprintf(stderr,"Error setting beginning time of the certificate\n");
        return -1;
     }
     if (!(X509_gmtime_adj(X509_get_notAfter(cert),CERT_SECS))) {
        fprintf(stderr,"Error setting ending time of the certificate\n");
        return -1;
     }

     /* add x509v3 extensions as specified */
     X509V3_set_ctx (&ctx, CAcert, cert, NULL, NULL, 0);
     for (i = 0; i < EXT_COUNT; i++) {
         X509_EXTENSION *ext;
         if (!(ext = X509V3_EXT_conf(NULL, &ctx,
				   ext_ent[i].key, ext_ent[i].value))) {
	     fprintf(stderr,"Error on \"%s = %s\"\n",
		     ext_ent[i].key, ext_ent[i].value);
	     fprintf(stderr,"Error creating X509 extension object\n");
             return -1;
	 }
         if (!X509_add_ext(cert, ext, -1)) {
	    fprintf(stderr,"Error on \"%s = %s\"\n",
		    ext_ent[i].key, ext_ent[i].value);
	    fprintf(stderr,"Error adding X509 extension to certificate\n");
            return -1;
	 }
         X509_EXTENSION_free(ext);
     } /* for each extension */

     /* add the subjectAltName in the request to the cert */
     if (!X509_add_ext(cert, subjAltName, -1)) {
        fprintf(stderr,"Error adding subjectAltName to certificate\n");
        return -1;
     }

     /* sign the certificate with the CA private key */
     digest = EVP_sha1 ();
     if (!(X509_sign(cert, CApkey, digest))) {
        fprintf(stderr,"Error signing certificate\n");
        return -1;
     }

     /* write the completed certificate to the same file storing our key */
     if (!(fp = fopen(certfile,"a"))) {
        fprintf(stderr,"Error writing to certificate file\n");
        return -1;
     }
     if (PEM_write_X509(fp, cert) != 1) {
        fprintf(stderr,"Error while writing certificate\n");
        fclose(fp);
        return -1;
     }

#if 0
     /* append the root certificate to the same file storing our key
        and our own certificate that way the certificate chain 
        is complete? Not apparently needed */
     if (PEM_write_X509(fp,CAcert) != 1) {
        fprintf(stderr,"Error while appending root certificate\n");
        fclose(fp);
        return -1;
     }
#endif

     fclose(fp);

     return 1;
}

/* ************************************************************* */
/* initialize a SSL context; library has already been initialized 
   keyfile contains our key and valid certificate for use in
   SSL connections
*/

int initialize_ssl_ctx(char *keyfile)
{
     SSL_METHOD *method;

     /* sigpipe handler already setup by daemon */

     DBG1("initialize_ssl_ctx: using '%s'\n",keyfile);

     /* pick which SSL version we will use 
        method = SSLv3_method();
        method = TLSv1_method();

        Java apparently does not support SSLv3, so use
        SSLv23 to bend our requirements a bit; we still end up with
        a TLSv1 connection though

        Re the book, we should use SSLv23_method and then set
        an option in the SSL_CTX to remove SSLv2 from acceptable
        protocols.  Use SSL_OP_NO_SSLv2 to prevent SSLv2 from being used.
     */

     method = SSLv23_method();
     ctx = SSL_CTX_new(method);

     /* eliminate SSLv2 from acceptable protocol list */
     SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2);

     /* set our certificate and private key */
     if (SSL_CTX_use_certificate_chain_file(ctx,keyfile) < 1) {
        fprintf(stderr,"Failed to load cert file, %ld\n",
                ERR_get_error());
        ERR_print_errors_fp(stderr);
        return -1;
     }
     if (SSL_CTX_use_PrivateKey_file(ctx,keyfile,SSL_FILETYPE_PEM) < 1) {
        fprintf(stderr,"Failed to load private key file, %ld\n",
                ERR_get_error());
        return -1;
     }

     /* load our CA for evaluating certs */
     if (SSL_CTX_load_verify_locations(ctx,"cartographer.pem",NULL) < 1) {
        fprintf(stderr,"Failed to load CA file, %ld\n",
                ERR_get_error());
        return -1;
     }

#if 0
     /* dont bother us with retries -- rdk */
     SSL_CTX_set_mode(ctx,SSL_MODE_AUTO_RETRY);
#endif

     /* set the verify/authentication preferences - clients
        and servers have to authenticate with certs 

     SSL_CTX_set_verify(ctx,SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,0);

     */

     /* given how ugly the java code is to automatically create
        and sign certs, we should set the peer verification
        to be optional; when agents talk to each other,
        they definitely should be authenticated but when
        a java manager queries, let them slide only if
        read/only queries.  Set a flag in the session that is 
        if the client has been authenticated
          -- rdk 2/14/2008 */

     SSL_CTX_set_verify(ctx,SSL_VERIFY_PEER,0);

     /* set our SSL context session caching mode; turn off, so we
        dont cache sessions and info */
     SSL_CTX_set_session_cache_mode(ctx,SSL_SESS_CACHE_OFF);

     return 1;
}

/* ************************************************************* */
/* check a certificate for validity; return -1 if error, 1 if OK
   and return 0 if no peer certificate; if we verify a certificate
   we set the authen flag for the session */

int check_certificate(xmpSession_t *sp)
{
     SSL *ssl;
     X509 *peerCert;
     char peer_name[512];
     int extcount,i;
     X509_EXTENSION *ext;
     const char *extstr;

     if (sp == NULL) {
        return -1;
     }

     ssl = (SSL *)sp->ssl;

     /* check verification */
     if (SSL_get_verify_result(ssl) != X509_V_OK) {
        fprintf(stderr,"Certificate does not verify\n");
        return -1;
     }

     /* check verification succeeded or no peer certificate was presented */

     if ((peerCert = SSL_get_peer_certificate(ssl)) == NULL) {
        /* no client certificate; thats OK, but we dont set the authen
           bit in flags */
        DBG("check_certificate: no peer certificate\n");
        return 0;
     }

     DBG("check_certificate: cert present, checking\n");

     X509_NAME_get_text_by_NID(X509_get_subject_name(peerCert),
                              NID_commonName,peer_name,512);

     DBG2("check_cert: remote IP:%x %s\n",sp->remIp,peer_name);

     /* check subjectAltName for dns name */
     if ((extcount = X509_get_ext_count(peerCert)) > 0) {

        DBG1("check_cert: %d extensions\n",extcount);

        for (i=0; i<extcount; i++) {
            
	    ext = X509_get_ext(peerCert,i);
            extstr = OBJ_nid2sn(OBJ_obj2nid(X509_EXTENSION_get_object(ext)));

            if (strcmp(extstr,"subjectAltName") == 0) {

	       DBG("check_certificate: found subjectAltName\n");


            } /* if subjectAltName */

        } /* for each extension */
     }
     else {
        fprintf(stderr,"check_certificate: peer cert didnt have extensions\n");
     }

     X509_free(peerCert);

     /* here we do DNS to resolve that the IP address we are connected to 
        matches the name in the certificate XXX */

     sp->flags |= XMP_SESSION_AUTHEN;

     DBG("check_certificate: returning authenticated\n");

     return 1;
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* call thread setup prior to any threads and ssl openssl calls */

static pthread_mutex_t *mutex_buf = NULL;

#ifndef WIN32
static unsigned long id_function() { return (unsigned long)pthread_self(); }
#else
static unsigned long id_function() { return (unsigned long)GetCurrentThreadId(); }
#endif

static void locking_function(int mode, int n, const char *file, int line)
{
     if (mode & CRYPTO_LOCK)
        pthread_mutex_lock(&mutex_buf[n]);
     else
        pthread_mutex_unlock(&mutex_buf[n]);
}

int THREAD_setup() 
{
     int i;

     mutex_buf = (pthread_mutex_t *)malloc(CRYPTO_num_locks() * 
                                           sizeof(pthread_mutex_t));
     if (mutex_buf == NULL) {
        fprintf(stderr,"malloc of mutex space failed\n");
        return -1;
     }
     for (i=0; i<CRYPTO_num_locks(); i++) {
         pthread_mutex_init(&mutex_buf[i],NULL);
     }
     CRYPTO_set_id_callback(id_function);
     CRYPTO_set_locking_callback(locking_function);

     return 1;
}

/* call thread cleanup when all threads are finished (exit) */

int THREAD_cleanup()
{
     int i;

     if (mutex_buf == NULL)
        return 0;
  
     CRYPTO_set_id_callback(NULL);
     CRYPTO_set_locking_callback(NULL);
     for (i=0; i<CRYPTO_num_locks(); i++) {
         pthread_mutex_destroy(&mutex_buf[i]);
     }
     free(mutex_buf);
     mutex_buf = NULL;

     return 1;
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */

int get_fqdn(char *fqdn, int maxlen)
{
     char hostname[128], domainname[128];

     if ((fqdn == NULL) || (maxlen < 32)) {
        fprintf(stderr,"get_fqdn: invalid args\n");
        return -1;
     }
     fqdn[0] = 0;

     if (gethostname(hostname,127) < 0) {
        return -1;
     }

     domainname[0] = 0;
     getdomainname(domainname,127);
     if (strcmp(domainname,"(none)") == 0)
        domainname[0] = 0;

     DBG2("Found hostname '%s' domainname '%s'\n",hostname,domainname);

     if (strstr(hostname,domainname) == NULL) {
        sprintf(fqdn,"%s.%s",hostname,domainname);
     }
     else {
        sprintf(fqdn,"%s",hostname);
     }
     
     return 1;
}

/* ************************************************************* */

int seed_prng()
{
#ifdef WIN32
     RAND_screen();
#else
     RAND_load_file("/dev/urandom",64);
#endif

     return 1;
}

/* ************************************************************* */
/* make our (agent's xmpd) private key and write to 
   file specified by caller (e.g. xmpd.pem) within cartographer's 
   home directory 
*/

int make_private_key(char *certfile)
{
     RSA *rsa;
     int ret;
     BIGNUM *bn;
     FILE *fp;

     if ((certfile == NULL) || (certfile[0] == 0)) {
        fprintf(stderr,"No cert file\n");
        return -1;
     }

     /* RSA_generate_key_ex or RSA_generate_key */
     /* ret=1 means succeed */
     rsa = RSA_new();
     bn = BN_new();
     BN_set_word(bn, 0x10001);

     if ((ret = RSA_generate_key_ex(rsa, 1024, bn, NULL)) < 1) {
        printf("make_private_key: key generation returned %d, %ld\n",ret,
               ERR_get_error());
     }
     else {
        /* have a private key pair, what do I do now ? XXX */
        DBG1("make_private_key: succeeded, returned %d\n",ret);
        DBG1("make_private_key: check key returned %d\n",
              RSA_check_key(rsa));

        /* write to temp file erasing anything that might be in it */
        if ((fp = fopen(certfile,"w")) != NULL) {
	   PEM_write_RSAPrivateKey(fp,rsa,NULL,NULL,0,NULL,NULL);
	   fclose(fp);
        }
     }

     /* cleanup! */
     if (bn) BN_free(bn);
     if (rsa) RSA_free(rsa);

     return 1;
}

/* ************************************************************* */
/* this is our chance to initialize any SSL related stuff
   that we need for secure communications (all platforms)
   return 0 failure, 1 success

   Each Cartographer agent is its own signing authority (CA) using
   the same private key and cert; this info is stored in 
   cartographer.pem and is distributed with Cartographer
 
   Each agent also needs its own key and cert; that is stored
   in xmpd.pem
   private key is written into xmpd.pem
   cert is then appended to xmpd.pem 

*/

int init_ssl_stuff(char *fqdn, char *certfile, char *csrfile)
{
     FILE *fp;

     if ((fqdn == NULL) || (fqdn[0] == 0)) {
        fprintf(stderr,"init_ssl_stuff: no FQDN\n");
        return -1;
     }
     if ((certfile == NULL) || (certfile[0] == 0)) {
        fprintf(stderr,"init_ssl_stuff: no certfile\n");
        return -1;
     }
     if ((csrfile == NULL) || (csrfile[0] == 0)) {
        fprintf(stderr,"init_ssl_stuff: no csrfile\n");
        return -1;
     }

     THREAD_setup();

     SSL_library_init();
     SSL_load_error_strings();

     /* create and error write context? */

     /* seed the PRNG first! if caller already called ssl init, then
        random number generate should have been seeded */

     /* if we seed the PRNG, we pollute the data read out of SSL
        as far as valgrind is concerned; consequently, when we're
        diagnosing memory errors, dont seed it; also, we're using
        a libssl/libcrypto compiled with -DPURIFY */

#if 1
     seed_prng();
#endif

     /* load our CA information; we are our own/only CA and
        if we cant load the info, we cannot securely communicate.
        stored in cartographer.pem -> key and cert
      */
     if ((fp = fopen("cartographer.pem","r")) == NULL) {
        fprintf(stderr,"Cannot find cartographer.pem\n");
        return -1;
     }
     fclose(fp);

     DBG("init_ssl_stuff: found cartographer.pem\n");

     /* load our keys and certificates; if we cant load them,
        this is our first time running SSL-enabled and we
        need to go create them */
     if ((fp = fopen(certfile,"r")) == NULL) {

        /* create our private key and store in xmpd.pem */
        if (make_private_key(certfile) < 0) {
	   return -1;
        }

        /* create our CSR */
        if (create_cert_request(fqdn,certfile,csrfile) < 0) {
	   return -1;
        }

        DBG("init_ssl_stuff: created csr\n");

        /* sign the CSR and create our cert
           and store both in that specified by caller (xmpd.pem) */
        if (generate_cert(certfile,csrfile) < 0) {
	   return -1;
        }

        DBG("init_ssl: generated cert\n");

        /* set file permissions accordingly */

     }
     else {
        DBG("init_ssl_stuff: found private key and cert\n");
        fclose(fp);
     }

     /* initialize our SSL context */
     initialize_ssl_ctx(certfile);

     return 1;
}

/* ************************************************************* */

int terminate_ssl_stuff()
{
     THREAD_cleanup();
     return 0;
}

/* ************************************************************* */
/* read a RSA pub/priv key from PEM file and return it
   in ptr, null otherwise */

RSA *read_rsa_key_from_pem_file(char *filename)
{
     RSA *rsa;
     FILE *fp;

     if ((rsa = RSA_new()) == NULL) {
        fprintf(stderr,"Unable to alloc RSA\n");
        return NULL;
     }

     if ((filename == NULL) || (filename[0] == 0)) {
        fprintf(stderr,"Cant read rsa from null file\n");
        return NULL;
     }

     if ((fp = fopen(filename,"r")) == NULL) {
        return NULL;
     }

     PEM_read_RSAPrivateKey(fp,&rsa,NULL,NULL);

     fclose(fp);
   
     return rsa;
}

/* ************************************************************* */
/* discover where we live (assuming a full pathname was used to 
   invoke us)

   UNIX:

   getcwd() returns /, /etc/init.d/, or /etc/rcX.d if
   run from an init script; check to see if 
   we can find out where the binary actually lives by
   checking argv[0] by checking for an absolute path

   WIN32:

   on windows, _getcwd() returns c:\windows\system32
   if we are running as a service; consequently we
   need to find out where we live by looking at
   command-line assuming its a full pathname 

   GetCommandLine can return a space-separated string delimited
   by quotes; if this is the case, make sure we get rid
   of the leading quote; the trailing quote will have been
   chopped off

*/

#ifdef WIN32

int get_home_dir(char *myname, char *dirname, int maxlen)
{
     char pathname[1024], *cstr, *cptr;
     char program[128];

     /* just in case */
     pathname[0] = 0;          

     if (_getcwd(pathname,1023) == NULL) {
        return -1;
     }

     cstr = GetCommandLine();

     DBG2("get_home_dir: cwd:'%s' cmdline is '%s'\n",pathname,cstr);

     /* find the last backslash */
     sprintf(program,"\\%s",myname);
     cptr = strstr(cstr,program);
     if (cptr != NULL) {
        *cptr = 0; 
     }
     if (*cstr == '\"')
        cstr++;

     snprintf(dirname,maxlen,"%s",cstr);
     dirname[maxlen-1] = 0;

     DBG1("get_home_dir: I live at '%s'\n",dirname); 

     return 1;
}

#endif /* WIN32 */

#ifndef WIN32

int get_home_dir(char *myname, char *dirname, int maxlen)
{
     return -1;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* from the dreaded applink.c */

#ifdef WIN32

#define APPLINK_STDIN	1
#define APPLINK_STDOUT	2
#define APPLINK_STDERR	3
#define APPLINK_FPRINTF	4
#define APPLINK_FGETS	5
#define APPLINK_FREAD	6
#define APPLINK_FWRITE	7
#define APPLINK_FSETMOD	8
#define APPLINK_FEOF	9
#define APPLINK_FCLOSE 	10	/* should not be used */

#define APPLINK_FOPEN	11	/* solely for completeness */
#define APPLINK_FSEEK	12
#define APPLINK_FTELL	13
#define APPLINK_FFLUSH	14
#define APPLINK_FERROR	15
#define APPLINK_CLEARERR 16
#define APPLINK_FILENO	17	/* to be used with below */

#define APPLINK_OPEN	18	/* formally can't be used, as flags can vary */
#define APPLINK_READ	19
#define APPLINK_WRITE	20
#define APPLINK_LSEEK	21
#define APPLINK_CLOSE	22
#define APPLINK_MAX	22	/* always same as last macro */

#ifndef APPMACROS_ONLY
#include <stdio.h>
#include <io.h>
#include <fcntl.h>

static void *app_stdin(void)		{ return stdin;  }
static void *app_stdout(void)		{ return stdout; }
static void *app_stderr(void)		{ return stderr; }
static int   app_feof(FILE *fp)		{ return feof(fp); }
static int   app_ferror(FILE *fp)	{ return ferror(fp); }
static void  app_clearerr(FILE *fp)	{ clearerr(fp); }
static int   app_fileno(FILE *fp)	{ return _fileno(fp); }
static int   app_fsetmod(FILE *fp,char mod)
{ return _setmode (_fileno(fp),mod=='b'?_O_BINARY:_O_TEXT); }

#ifdef __cplusplus
extern "C" {
#endif

__declspec(dllexport)
void **
#if defined(__BORLANDC__)
__stdcall	/* __stdcall appears to be the only way to get the name
		 * decoration right with Borland C. Otherwise it works
		 * purely incidentally, as we pass no parameters. */
#else
__cdecl
#endif
OPENSSL_Applink(void)
{ static int once=1;
  static void *OPENSSL_ApplinkTable[APPLINK_MAX+1]={(void *)APPLINK_MAX};

    if (once)
    {	OPENSSL_ApplinkTable[APPLINK_STDIN]	= app_stdin;
	OPENSSL_ApplinkTable[APPLINK_STDOUT]	= app_stdout;
	OPENSSL_ApplinkTable[APPLINK_STDERR]	= app_stderr;
	OPENSSL_ApplinkTable[APPLINK_FPRINTF]	= fprintf;
	OPENSSL_ApplinkTable[APPLINK_FGETS]	= fgets;
	OPENSSL_ApplinkTable[APPLINK_FREAD]	= fread;
	OPENSSL_ApplinkTable[APPLINK_FWRITE]	= fwrite;
	OPENSSL_ApplinkTable[APPLINK_FSETMOD]	= app_fsetmod;
	OPENSSL_ApplinkTable[APPLINK_FEOF]	= app_feof;
	OPENSSL_ApplinkTable[APPLINK_FCLOSE]	= fclose;

	OPENSSL_ApplinkTable[APPLINK_FOPEN]	= fopen;
	OPENSSL_ApplinkTable[APPLINK_FSEEK]	= fseek;
	OPENSSL_ApplinkTable[APPLINK_FTELL]	= ftell;
	OPENSSL_ApplinkTable[APPLINK_FFLUSH]	= fflush;
	OPENSSL_ApplinkTable[APPLINK_FERROR]	= app_ferror;
	OPENSSL_ApplinkTable[APPLINK_CLEARERR]	= app_clearerr;
	OPENSSL_ApplinkTable[APPLINK_FILENO]	= app_fileno;

	OPENSSL_ApplinkTable[APPLINK_OPEN]	= _open;
	OPENSSL_ApplinkTable[APPLINK_READ]	= _read;
	OPENSSL_ApplinkTable[APPLINK_WRITE]	= _write;
	OPENSSL_ApplinkTable[APPLINK_LSEEK]	= _lseek;
	OPENSSL_ApplinkTable[APPLINK_CLOSE]	= _close;

	once = 0;
    }

  return OPENSSL_ApplinkTable;
}

#ifdef __cplusplus
}
#endif
#endif


#endif /* WIN32 */
