/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmpd.c */
/* rdk@krupczak.org */

/* daemon engine and common routines */

#ifdef WIN32
#include <windows.h>
#include <direct.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <errno.h>

#ifndef WIN32
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/utsname.h>
#include <sys/resource.h>
#include <strings.h>
#include <arpa/inet.h>
#endif

#include <string.h>
#include <stdlib.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

#include "globals.h"
#include "xmp.h"
#include "socklib.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "xmp_events.h"
#include "modules.h"
#include "plugins.h"
#include "xmp_users.h"

/* ************************************************************* */

extern xmpStats_t xmpStats;

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* scalar method routines */

static void inGetRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;

     if (xmpd_verbose)
        printf("GetRequest starting\n"); 

     DBG("inGetRequest starting\n"); 

     /* first copy PDU and variables */
     copyXmpPdu(in,out);
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->messageID = in->messageID;
     out->msgType = Response;

     DBG2("inGetRequest: message IDs are in:%lld out:%lld\n",
	  in->messageID,out->messageID);

     varptr = out->varList;

     i = 0;
     while (varptr != NULL) {

       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);

       if (varentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          break;
       }

       /* check if entry has function-pointer for get */
       if (varentry->get == NULL) {
 	  out->errorStatus = noAccess;
          out->errorIndex = i;
          break;
       }

       /* protect ourselves; variable should not be tabular */
       if (varentry->tableName[0] != 0) {
	  out->errorStatus = noAccess;
          out->errorIndex = i;
       }

       /* if GetRequest had nullSyntax, set the outgoing syntax */
       if (varptr->varSyntax == nullSyntax) {
 	  varptr->varSyntax = varentry->syntax;
       }

       /* like SNMP, on Gets, we essentially ignore the incoming 
          syntax and value */

       if (xmpd_verbose)
          printf("GetRequest going to get for %s\n",varptr->varName.objName); 

       /* fetch the value */
       e = (varentry->get)(varentry,varptr);
       if (e != noError) {
	  out->errorStatus = e;
          out->errorIndex = i;
       }
       DBG1("GetRequest got %s!\n",varptr->varName.objName); 

       varptr = varptr->next;
       i++;
     }

     if (xmpd_verbose)
        printf("inGetRequest: finished, %d vars\n",i);

     return; 
}

/* ************************************************************* */
/* scalar set request
   engine first calls get method to see if variable is available;
   if variable available, then set is called in two pass;
   first pass tests to see if value is set'able; 2nd pass,
   performs the set.  All variables in Set PDU are all un-related as 
   far as the protocol and engine goes 
*/

static void inSetRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{ 
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;

     /* prepare outgoing PDU by copying it and variables */
     copyXmpPdu(in,out);
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     /* verify that no table variables and do initial get to test 
        and make sure that everything is set'able */
     varptr = out->varList;
     i = 0;
     while (varptr != NULL) {

       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);

       if (varentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          return;
       }
       if ((varentry->get == NULL) || (varentry->set == NULL)) {
 	  out->errorStatus = noAccess;
          out->errorIndex = i;
          return;
       }
       if (varentry->tableName[0] != 0) {
	  out->errorStatus = noAccess;
          out->errorIndex = i;
          return;
       }
       if (varptr->varSyntax != varentry->syntax) {
	  printf("inSetRequest: syntax does not match for %d\n",i);
	  out->errorStatus = badValue;
          out->errorIndex = i;
          return;
       }
       e = (varentry->get)(varentry,varptr);
       if (e != noError) {
	  out->errorStatus = e;
          out->errorIndex = i;
          return;
       }

       varptr = varptr->next;
       i++;
     }

     /* we got this far; gets all worked, everything is writable */
     /* since we did a get, trash the out PDU and re-copy/init it */
     xmpPduCleanup(out);
     copyXmpPdu(in,out);
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;

     /* first pass tests the set */
     varptr = out->varList;
     i = 0;
     while (varptr != NULL) {

       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);

       e = (varentry->set)(varentry,varptr,0);
       if (e != noError) {
	  out->errorStatus = e;
          out->errorIndex = i;
          return;
       }

       varptr = varptr->next;
       i++;
     }

     /* second pass performs the set */
     varptr = out->varList;
     i = 0;
     while (varptr != NULL) {

       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);

       e = (varentry->set)(varentry,varptr,1);
       if (e != noError) {
	  out->errorStatus = e;
          out->errorIndex = i;
          return;
       }

       varptr = varptr->next;
       i++;
     }

     if (xmpd_verbose)
        printf("inSetRequest: finished, %d vars\n",i);

     return; 
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* misc PDU types */

/* need to match response with appropriate session */

static void inResponse(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     fprintf(stderr,"Response not yet supported\n"); 

     copyXmpPdu(in,out);
     out->errorStatus = genError;
     out->messageID = in->messageID;

     return; 
}

static void inTrap(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out, 
                   u_char *buffer, int len)
{ 
     xmpErrorStatus_t e;
     char ipAddressStr[128];

     copyXmpPdu(in,out);

     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     /* for now, we are going to store traps in our trap table 
        where we they will reside until we age them out */

     convert_ipaddr(sp->remIp,ipAddressStr,127);

     if ((e = receiveTrap(ipAddressStr,in,buffer,len)) != noError) {
        out->errorStatus = e;
     }

     /* trap sender gets a response acknowledging the trap */

     return; 
}

static void inInformation(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out, 
                          u_char *buffer, int len)
{
     xmpErrorStatus_t e;
     char ipAddressStr[128];

     copyXmpPdu(in,out);

     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     convert_ipaddr(sp->remIp,ipAddressStr,127);

     if ((e = receiveInformation(ipAddressStr,in,buffer,len)) != noError) {
        out->errorStatus = e;
     }

     /* Information sender gets a standard response pdu */

     return; 
}

/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* table operations */

static void inInsertRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;
     xmpVar_t *tableentry;

     copyXmpPdu(in,out);
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     /* see if table exists and has insert function */
     tableentry = xmpFindVar(out->tableMib,out->tableName);
     if (tableentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = 0;
          return;
     }
     if (tableentry->insert == NULL) {
        out->errorStatus = noAccess;
        out->errorIndex = 0;
        return;
     }

     varptr = out->varList;
     /* first, verify if everything is tabular and in same table */
     i = 0;
     while (varptr != NULL) {
       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);
     
       /* each variable in PDU must be within table */
       if (varentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          return;
       }

       if (varentry->tableName[0] == 0) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          return;
       }

       /* verify that columns are in targeted table */
       if ((strcmp(out->tableName,varentry->tableName) != 0) ||
           (strcmp(out->tableMib,varentry->mibName) != 0)) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          fprintf(stderr,"%s is not in table %s\n",
                  varentry->objName,out->tableName);
          return;
       }

       varptr->entryPtr = (void *)varentry; /* save it for method routine!! */

       varptr = varptr->next;
       i++;
     }

     /* call method routine now that we have verified everything */

     /* method routine sets errorIndex if non-zero, and creates
        table entry if everything checks out */

     e = (tableentry->insert)(out->tableKey,out);
     out->errorStatus = e;

     if (xmpd_verbose)
        printf("inInsertRequest: finished, %d vars\n",out->numberVariables);
     if (xmpd_verbose)
        printf("inInsertRequest: finished, %d vars\n",out->numberVariables);

     return; 
}

/* ************************************************************* */

static void inDeleteRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;
     xmpVar_t *tableentry;

     copyXmpPdu(in,out);
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     /* see if table exists and has insert function */
     tableentry = xmpFindVar(out->tableMib,out->tableName);
     if (tableentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = 0;
          return;
     }
     if (tableentry->delete == NULL) {
        out->errorStatus = noAccess;
        out->errorIndex = 0;
        return;
     }

     /* now, to delete, one need only supply the key; however,
        if message contains columns, we need to insure that 
        all the columns are in the targettable */

     /* first, verify if everything is tabular and in same table */
     varptr = out->varList;
     i = 0;
     while (varptr != NULL) {
       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);
     
       /* each variable in PDU must be within table */
       if (varentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          return;
       }

       if (varentry->tableName[0] == 0) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          return;
       }

       /* verify that columns are in targeted table */
       if ((strcmp(out->tableName,varentry->tableName) != 0) ||
           (strcmp(out->tableMib,varentry->mibName) != 0)) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          fprintf(stderr,"%s is not in table %s\n",
                  varentry->objName,out->tableName);
          return;
       }

       varptr->entryPtr = (void *)varentry; /* save it for method routine!! */

       varptr = varptr->next;
       i++;
     }

     /* method routine sets errorIndex if non-zero, and deletes
        table entry if everything checks out; should the delete
        operation return the row(s) that were deleted? XXX */

     e = (tableentry->delete)(out->tableKey,out);
     out->errorStatus = e;

     if (xmpd_verbose)
        printf("inDeleteRequest: finished, %d vars\n",out->numberVariables);

     return; 
}

/* ************************************************************* */

static void inUpdateRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;
     xmpVar_t *tableentry;

     copyXmpPdu(in,out);  
     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->messageID = in->messageID;

     /* see if table exists and has insert function */
     tableentry = xmpFindVar(out->tableMib,out->tableName);
     if (tableentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = 0;
          return;
     }
     if (tableentry->delete == NULL) {
        out->errorStatus = noAccess;
        out->errorIndex = 0;
        return;
     }

     varptr = out->varList;
     /* first, verify if everything is tabular and in same table */
     i = 0;
     while (varptr != NULL) {
       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);
     
       /* each variable in PDU must be within table */
       if (varentry == NULL) {
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          return;
       }

       if (varentry->tableName[0] == 0) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          return;
       }

       /* verify that columns are in targeted table */
       if ((strcmp(out->tableName,varentry->tableName) != 0) ||
           (strcmp(out->tableMib,varentry->mibName) != 0)) {
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          fprintf(stderr,"%s is not in table %s\n",
                  varentry->objName,out->tableName);
          return;
       }

       varptr->entryPtr = (void *)varentry; /* save it for method routine!! */

       varptr = varptr->next;
       i++;
     }

     /* call method routine now that we have verified everything */

     /* method routine sets errorIndex if non-zero, and modifies
        table entry if everything checks out; if target row
        does not exist, noSuch is returned */

     e = (tableentry->update)(out->tableKey,out);
     out->errorStatus = e;

     if (xmpd_verbose)
        printf("inUpdateRequest: finished, %d vars\n",out->numberVariables);

     return; 
}

/* ************************************************************* */

static void inSelectRequest(xmpSession_t *sp, xmpPdu_t *in, xmpPdu_t *out)
{
     xmpVariable_t *varptr;
     xmpVar_t *varentry;
     xmpErrorStatus_t e;
     int i;
     xmpVar_t *tableentry;

     if (xmpd_verbose)
        printf("selectRequest starting\n"); 

     /* dont copy PDU variables since method routine will
        be putting its own vars */
   
     /* if there is something like a nosuch, then nothing
        gets copied and the response error nosuch has no 
        vars XXX */

     bcopy(in,out,sizeof(xmpPdu_t));

     out->errorStatus = noError;
     out->errorIndex = 0;
     out->msgType = Response;
     out->varList = NULL;
     out->varListTail = NULL;
     out->numberVariables = 0;
     out->messageID = in->messageID;

     DBG2("inSelectRequest: message IDs are in:%lld out:%lld\n",
          in->messageID,out->messageID);

     /* see if table exists */

     tableentry = xmpFindVar(out->tableMib,out->tableName);

     if (tableentry == NULL) {
          DBG2("inSelectRequest: couldnt find table %s.%s\n",
                out->tableMib,out->tableName);
 	  out->errorStatus = noSuchName;
          out->errorIndex = 0;
          xmpStats.outErrors++;
          return;
     }
     if (tableentry->select == NULL) {
        out->errorStatus = noAccess;
        out->errorIndex = 0;
        xmpStats.outErrors++;
        return;
     }

     varptr = in->varList;

     /* first, verify if everything is tabular and in same table */
     i = 0;
     while (varptr != NULL) {
       varentry = xmpFindVar(varptr->varName.mibName,
                             varptr->varName.objName);
     
       /* each variable in PDU must be within table */
       if (varentry == NULL) {
          DBG2("inSelectRequest: couldnt find table object %s.%s\n",
                varptr->varName.mibName,varptr->varName.objName);
	  copyXmpPdu(in,out);
 	  out->errorStatus = noSuchName;
          out->errorIndex = i;
          xmpStats.outErrors++;
          return;
       }

       if (varentry->tableName[0] == 0) {
	  copyXmpPdu(in,out);
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          xmpStats.outErrors++;
          return;
       }

       /* verify that columns are in targeted table */
       if ((strcmp(out->tableName,varentry->tableName) != 0) ||
           (strcmp(out->tableMib,varentry->mibName) != 0)) {
	  copyXmpPdu(in,out);
 	  out->errorStatus = notTabular;
          out->errorIndex = i;
          xmpStats.outErrors++;
          fprintf(stderr,"%s is not in table %s\n",
                  varentry->objName,out->tableName);
          return;
       }

       /* if GetRequest had nullSyntax, set the outgoing syntax */
       if (varptr->varSyntax == nullSyntax) {
 	  varptr->varSyntax = varentry->syntax;
       }
       varptr->entryPtr = (void *)varentry; /* save it for method routine!! */

       varptr = varptr->next;
       i++;
     }

     /* call the method routine now that we have verified
        everything */

     /* method routine sets errorStatus, errorIndex,
        numberVariables, mallocs/inits variableList,
        fills in values, etc. */

     e = (tableentry->select)(out->tableKey,in->varList,out);
     out->errorStatus = e;
 
     if (xmpd_verbose)
        printf("inSelectRequest: finished, %d vars\n",out->numberVariables);

     return; 
}

/* ************************************************************* */
/* 
   handle incoming sessions requesting management information; we
   use one thread per session

*/

void *xmpdInSessionHandler(void *ptr)
{
     int s; /* accepted socket */
     char ipAddressStr[128];
     int ret,replylen, need,sslret;
     xmpPdu_t pdu, replyPdu;
     int recvlen;
     xmpSession_t *sp;
     SSL *ssl;
     BIO *bio;
     extern SSL_CTX *ctx;

     sp = (xmpSession_t *)ptr;
     s = (int)sp->sockfd;
     xmpStats.inSessions++;
     sp->thr = pthread_self();
     sp->lastActivity = time(NULL);

     DBG2("xmpdInSessionHandler: starting with conn from %x:%d\n",
          sp->remIp,sp->remPort);

     /* at this point, we need to do an SSL accept and check
        the cert; do this prior to any more malloc'ing */
     ssl = SSL_new(ctx);
     bio = BIO_new_socket(sp->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,bio,bio);
     sp->ssl = (void *)ssl;
     sp->sbio = (void *)bio;

     if ((ret = SSL_accept(ssl)) < 0) {
        convert_ipaddr(sp->remIp,ipAddressStr,127);
        fprintf(stderr,"xmpdInSessionHandler: SSL accept failed\n");
        sslret = SSL_get_error(ssl,ret);
        fprintf(stderr,
                "xmpdInSessionHandler: SSL error %d from %s:%d, errno=%d\n",
                sslret,ipAddressStr,
                sp->remPort,
                errno);
        destroyXmpSession(sp);
        ERR_remove_state(0);
        pthread_exit(NULL);
     }

     /* check the client cert */
     if (check_certificate(sp) < 0) {
        fprintf(stderr,"xmpdInSessionHandler: check SSL cert failed\n");
        destroyXmpSession(sp);
        ERR_remove_state(0);
        pthread_exit(NULL);
     }

     DBG1("xmpdInSessionHandler: SSL ready to go with '%s'\n",
          SSL_get_version(ssl));

     /* clients do not have to authenticate; if they do authenticate
        set the authen bit in the session flags; check_cert() 
        sets this bit for us */
        
     /* higher-level authenticate user via appropriate means XXX */

     /* session now established; loop through reading and 
        replying to management messages; connection initiator
        severs connection/session when they finish */

     while (globalShutdown == 0) {

       /* check for inactivity; if no activity in XXX 
          seconds, close session */

       /* get the query in xml format */
       recvlen = xmpRecvMessage(sp);
       if (recvlen <= 0) {
	  break;
       }

       /* see if bzero'ing makes valgrind happy */
       bzero((void *)&pdu,sizeof(pdu));
       bzero((void *)&replyPdu,sizeof(replyPdu));

       /* parse the message */       
       if ((ret = parseXmpMessage(sp->recvBuf,recvlen,&pdu)) < 0) {

          convert_ipaddr(sp->remIp,ipAddressStr,127);          

	  fprintf(stderr,"sessionHandler: parse error from %s:%d\n",
                  ipAddressStr,sp->remPort);

          xmpStats.inXmlParseErrs++;

          /* send error message back */
          if (xmpSendErrorResponseFatal(sp,pdu.messageID,parseError,0) < 0) {
	     fprintf(stderr,
                     "inSessionHandler: error sending fatal error resp\n");
          }
          else {
	     fprintf(stderr,
                     "inSessionHandler: sent parseError reply\n");
          }

          continue; /* with while loop */
       }

       xmpStats.inTotalReqVars += pdu.numberVariables;

       /* now we have PDU in internal format; handle it */
       if ((xmpd_verbose) || (xmpd_trace)) {
 	  dumpInternalXmpMessage(&pdu);
       }

       /* check the message type versus the authenUser and
          see if sender has permission for this type of operation before
          we go through the hassle of doing all the work */
       if (validatePduPermissions(sp,&pdu) < 1) {
	  DBG4("inSessionHandler: invalid perms '%s'/%d from %x:%d\n",
               sp->authenUser,pdu.msgType,
               sp->remIp,sp->remPort);
          DBG1("inSessionHandler: mib:%s\n",
                pdu.varList->varName.mibName);

          /* generate an authorizationError error reply */
          xmpStats.inAuthenErrors++;

          /* send error message back */
          if (xmpSendErrorResponse(sp,&pdu,authorizationError,0) < 0) {
	     fprintf(stderr,
                     "inSessionHandler: error sending authen error resp\n");
          }
          else {
	     DBG1("inSessionHandler: sent authenError reply to %lx\n",
                  sp->remIp);
          }

          /* dont forget to clean up the original PDU!!! */
          xmpPduCleanup(&pdu);

          continue; /* with while loop */
       }

       switch (pdu.msgType) {

         /* scalars */
         case GetRequest:
	    xmpStats.inGets++;
	    inGetRequest(sp,&pdu,&replyPdu);
            break;
         case SetRequest:
	    /* sets require SSL authen bit to be set in addition to 
               the right protocol permissions with the authenUser */
	    xmpStats.inSets++;
	    inSetRequest(sp,&pdu,&replyPdu);
            break;

         /* misc */
         case Response:
	    xmpStats.inResponses++;
	    inResponse(sp,&pdu,&replyPdu);
            break;
         case Trap:
	    xmpStats.inTraps++;
	    inTrap(sp,&pdu,&replyPdu,sp->recvBuf,recvlen);
            break;
         case Information:
	    xmpStats.inInformation++;
	    inInformation(sp,&pdu,&replyPdu,sp->recvBuf,recvlen);
            break;

         /* table operations */
         case SelectTableRequest:
	    xmpStats.inSelects++;
	    inSelectRequest(sp,&pdu,&replyPdu);
            break;
         case InsertTableRequest:
	    xmpStats.inInserts++;
	    inInsertRequest(sp,&pdu,&replyPdu);
            break;
         case DeleteTableRequest:
	    /* deletes require authen bit to be set XXX */
	    xmpStats.inDeletes++; 
	    inDeleteRequest(sp,&pdu,&replyPdu);
            break;
         case UpdateTableRequest:
	    /* updates require authen bit to be set XXX */
	    xmpStats.inUpdates++;
	    inUpdateRequest(sp,&pdu,&replyPdu);
            break;

         default:
	    fprintf(stderr,"Message type %d currently not supported\n",
                    pdu.msgType);
            copyXmpPdu(&pdu,&replyPdu);
            replyPdu.errorStatus = genError;
       }

       /* re-encode reply PDU into XML */
       need = 0;

       replylen = buildXmpMessage(&replyPdu,sp->sendBuf,sp->sendBufSize,&need);

       if ((replylen < 0) && (need > 0)) {
	  sp->sendBuf = realloc(sp->sendBuf,need+2); 
          if (sp->sendBuf == NULL) {
	     fprintf(stderr,"unable to realloc send buf\n");
             destroyXmpSession(sp);
             ERR_remove_state(0);
             pthread_exit(NULL);
          }
	  sp->sendBufSize = need + 2;
          need = 0;
          replylen = buildXmpMessage(&replyPdu,sp->sendBuf,sp->sendBufSize,
                                     &need);
       }

       if (replylen < 0) {
 	  /* bump up stats XXX */
	  fprintf(stderr,"failed to build reply message\n");
       }
       else {
	  ret = xmpSendMessage(sp,sp->sendBuf,replylen);
       }

       xmpStats.outTotalReqVars += replyPdu.numberVariables;     

       /* cleanup internal PDU (e.g. varlist) */
       xmpPduCleanup(&pdu);
       xmpPduCleanup(&replyPdu);

       sp->lastActivity = time(NULL);

     } /* while messages in session and not shutdown */

     /* cleanup/destroy the session, close socket! */
     destroyXmpSession(sp);

     /* remove SSL error state for this thread */
     ERR_remove_state(0);

     if (xmpd_verbose)
       printf("SessionHandler: %d thread exiting\n",s);

     pthread_exit(NULL);
}

/* ************************************************************* */
/* listen for incoming connections requesting to initiate management 
   information exchange (peers, managers, etc.) 
*/

void *xmpdListenForConnections(void *arg)
{
     int s, s1,ret;
     u_int fromaddr;
     u_short fromport;
     pthread_t t;
     char ipstr[64];
     void *xmpdInSessionHandler(void *);
     xmpSession_t *sp;
     char message[128];

     if (xmpd_verbose)
        printf("xmpdListenForConnections: starting\n");

     /* get socket */
     if ((s = get_tcp_socket()) < 0) {
        perror("ListenForConn: failed to get socket\n");
        xmpd_log("listenForConnections: failed to get socket");
        sprintf(message,"Socket is %d",s);
        xmpd_log(message);
#ifdef WIN32
        printf("ListenForConn: winsock error %d\n",WSAGetLastError());
#endif
        return NULL;
     }

     /* bind socket to well-known port */
     if (bind_tcp_socket(s,0,(u_short)htons(xmpd_port),1) < 0) {
        perror("ListenForConn: bind failed");
        sprintf(message,"listenForConnections: bind failed,%d",errno);
        xmpd_log(message);
        return NULL;
     }

     /* listen for connections */
     if (listen_for_tcp_connection(s,5) < 0) {
        perror("Listen failed");
        sprintf(message,"listenForConnections: listen failed,%d",errno);
        xmpd_log(message);
        return NULL;
     }

     sprintf(message,"listenForConnections: main socket is %d",s);
     xmpd_log(message);

     while (globalShutdown == 0) {

       if (connection_pending(s,5) == 1) {

	  /* break from while loop if we are in midst of shutting down */
	  if (globalShutdown == 1)
	     break;

          /* accept connection, create thread to handle it */
          s1 = accept_tcp_connection_request(s, (u_int *)&fromaddr,
                                             (u_short *)&fromport);

          convert_ipaddr(fromaddr,ipstr,sizeof(ipstr));
          if (xmpd_verbose)
             fprintf(stderr,"Accepted connection from %x %s, port:%d\n",
		    fromaddr,
        	    ipstr,
                    ntohs(fromport));

          /* create a session struct; our session thread
             will complete the SSL portion of the connection */
          if ((sp = newXmpSession(fromaddr,fromport,s1,
		  XMP_SESSION_INCOMING|XMP_SESSION_CONNECTED,NULL)) == NULL) {
 	     fprintf(stderr,"failed to create session for incoming conn\n");
             xmpd_log("listenForConnections: failed to create session");
             continue;
          }

	  /* break from while loop if we are in midst of shutting down */
	  if (globalShutdown == 1) {
	     destroyXmpSession(sp);
	     break;
          }

          /* Very important!!!!
             Detach worker thread since we are not going to wait/join for 
             it to exit.  If we dont detach, when thread exits
             it will leave a whole pile of memory sitting waiting
             for us to join; I've estimated that memory to be about
             512K (as evidenced by ps).
          */

#ifndef WIN32
          pthread_attr_t attr;

          /* dont forget to initialize attributes! */
          pthread_attr_init(&attr);
          pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);

          if ((ret = pthread_create(&t,&attr,xmpdInSessionHandler,(void *)sp)) != 0) {
	     fprintf(stderr,"create of session thread failed, %d\n",ret);
             xmpd_log("listenForConnections: create session thread failed");
             destroyXmpSession(sp); /* !!! */
             continue; /* with while loop */
	  }

#if 0
          /* on older (2.4) systems, we seem to have problems with detach
             failures thus leaking memory; if we create the thread with
             the detach state, we then dont have to detach and hopefully
             we avoid memory leaking due to not joining them. */
          if ((ret = pthread_detach(t)) != 0) {
	     fprintf(stderr,"detach of session thread %lx failed, %d\n",t,ret);
             xmpd_log("listenForConnections: session thread detach failed");
          }

#endif 

#endif /* NOT WIN32 */

#ifdef WIN32
          /* in version 1.xx of pthreads, we cant see to get attributes
             correct due to syntax errors in the type declaration */
          if ((ret = pthread_create(&t,NULL,xmpdInSessionHandler,(void *)sp)) != 0) {
	     fprintf(stderr,"create of session thread failed, %d\n",ret);
             xmpd_log("listenForConnections: create session thread failed");
             destroyXmpSession(sp); /* !!! */
             continue; /* with while loop */
	  }
          if ((ret = pthread_detach(t)) != 0) {
	     fprintf(stderr,"detach of session thread %lx failed, %d\n",t,ret);
             xmpd_log("listenForConnections: session thread detach failed");
          }
#endif /* WIN32 */

       } /* if connection pending */

       /* continue waiting for connections unless global shutdown */

     } /* while */

     /* close socket just to make sure  . . . */
#ifndef WIN32
     close(s);
#else
     closesocket(s);
#endif

     if (xmpd_verbose)
        printf("xmpdListenForConnections: finishing\n");
     xmpd_log("xmpdListenForConnections finishing");

     return NULL;
}

/* ************************************************************* */

int usage(char *progname)
{
     fprintf(stderr,"usage: %s \n",progname);
     fprintf(stderr,"\tVersion: %3.2f\n", xmpd_version);
     fprintf(stderr,"\tWeb: http://www.krupczak.org/\n");
     fprintf(stderr,"\tEmail: info@krupczak.org\n");

     fprintf(stderr,"\n");

     fprintf(stderr,"\t-b\tbackground/daemon mode\n");
     fprintf(stderr,"\t-g\tdump UNIX core on crash\n");
     fprintf(stderr,"\t-t\t dump sent/received messages (trace)\n");
     fprintf(stderr,"\t-h\thelp (args)\n");
     fprintf(stderr,"\t-v\tverbose \n");
     fprintf(stderr,"\t-V\tprint out version info and exit\n");
     fprintf(stderr,"\t-l\tprint out license info and exit\n");
     fprintf(stderr,"\t-f FILE\tuse config-file FILE\n");
#ifdef WIN32
     fprintf(stderr,"\t-s\tdo not run as service\n");
#endif

     fprintf(stderr,"\n");

#ifdef LINUX
     fprintf(stderr,"\tLinux build\n");
#endif

#ifdef SOLARIS
#ifdef SOLX86
     fprintf(stderr,"\tSolarisX86 build\n");
#else
     fprintf(stderr,"\tSolaris build\n");
#endif
#endif

#ifdef XMPDEBUG
     fprintf(stderr,"\tXMPDEBUG enabled\n");
#endif

     fprintf(stderr,"\n");
     exit(0);
}

/* ************************************************************* */

int version_info()
{
     fprintf(stderr,"Cartographer version %3.2f\n",xmpd_version);
#ifdef LINUX
     fprintf(stderr,"\tLinux build\n");
#endif

#ifdef SOLARIS
#ifdef SOLX86
     fprintf(stderr,"\tSolarisX86 build\n");
#else
     fprintf(stderr,"\tSolaris build\n");
#endif
#endif
#ifdef WIN32
     fprintf(stderr,"\tWin32 build\n");
#endif

     /* build flags */
#ifdef XMPDEBUG
     fprintf(stderr,"XMPDEBUG enabled\n");
#else
     fprintf(stderr,"XMPDEBUG not enabled\n");
#endif

     /* plugin suffix */
     fprintf(stderr,"PLUGINSUFFIX='%s'\n",PLUGINSUFFIX);

     /* on win32, we want to know pthread version */
#ifdef WIN32
     fprintf(stderr,"pthreads version '%s'\n",PTW32_VERSION_STRING);
#endif /* WIN32 */


     fprintf(stderr,"\n");
     exit(0);
}

int license_info()
{

#define LICENSE_INFO "\n COPYRIGHT 2008, KRUPCZAK.ORG, LLC.\n \
\n \
THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY\n \
APPLICABLE LAW. KRUPCZAK.ORG, LLC PROVIDES THE PROGRAM 'AS IS'\n \
WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,\n \
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND\n \
FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY\n \
AND PERFORMANCE OF THE PROGRAM IS WITH THE USER. SHOULD THE PROGRAM\n \
PROVE DEFECTIVE, THE USER ASSUMES THE COST OF ALL NECESSARY\n \
SERVICING, REPAIR OR CORRECTION.\n \
\n \
IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n \
WILL KRUPCZAK.ORG, LLC, OR ANY OTHER PARTY WHO CONVEYS THE PROGRAM,\n \
BE LIABLE TO THE USER FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL,\n \
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR\n \
INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF\n \
DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY THE\n \
USER OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY\n \
OTHER PROGRAMS), EVEN IF KRUPCZAK.ORG, LLC OR OTHER PARTY HAS BEEN\n \
ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.\n"

     fprintf(stderr,"%s\n",LICENSE_INFO);

     exit(0);
}

/* ************************************************************* */

int init_xmpd(int argc, char *argv[])
{
     extern struct utsname systemuts;
     extern time_t xmpd_init_time;
     void *houseKeeping();
     char *cstr;
     pthread_t cl,hk;
     int readXmpdConfigFile(), initUsers();
     int initVars(), intCore(),i, make_a_daemon_out_of_me(), mini_daemon();
     char logBuf[1024], *cptr;
     char appname[64];

#ifdef WIN32
     WSADATA wsaData;

     if (WSAStartup(0x202,&wsaData) == SOCKET_ERROR) {
        perror("WSAStartup failed");
        WSACleanup();
        if (win32service)
 	   return -1;
        else 
           exit(-1);
     }
#endif

#ifndef WIN32
     if (dumpcore) {
        struct rlimit rl;
        bzero((void *)&rl,sizeof(rl));
        rl.rlim_cur = RLIM_INFINITY;
        rl.rlim_max = RLIM_INFINITY;
        if (setrlimit(RLIMIT_CORE,&rl) < 0) {
	   fprintf(stderr,"xmpd: unable to set core size resource limit, %d\n",
                   errno);
        }
     }
#endif

     /* figure out what kind of system we are running on */
     uname(&systemuts);

     xmpd_init_time = time(NULL);

     /* before we daemonize and chdir, figure out where we
        live and squirl that away; that location is where we will
        find plugins, data files, arch-dependent files, etc. */

#ifndef WIN32
     if (getcwd(xmpd_home_dir,FILENAME_MAX-1) == NULL) {
        fprintf(stderr,"xmpd: home directory is too large for buffer\n");
        if (win32service)
	   return -1;
        else
           exit(-1);
     }

     /* getcwd() returns /, /etc/init.d/, or /etc/rcX.d if
        run from an init script; check to see if 
        we can find out where the binary actually lives by
        checking argv[0] by checking for an absolute path */
     getApplicationName(appname,sizeof(appname));

     if ((cptr = strstr(argv[0],appname)) != NULL) {
        *cptr = 0;
        if (argv[0][0] == '/')
           sprintf(xmpd_home_dir,"%s",argv[0]);
     }

#else /* WIN32 */

     /* on windows, _getcwd() returns c:\windows\system32
        if we are running as a service; consequently we
        need to find out where we live by looking at
        command-line assuming its a full pathname 
      */

     if (_getcwd(xmpd_home_dir,FILENAME_MAX-1) == NULL) {
        fprintf(stderr,"xmpd: home directory is too large for buffer\n");
        if (win32service)
	   return -1;
        else
           exit(-1);
     }
     if (win32service) {

        cstr = GetCommandLine();
        /* GetCommandLine can return a space-separated string delimited
           by quotes; if this is the case, make sure we get rid
           of the leading quote; the trailing quote will have been
           chopped off
	*/
           
       if ((cptr = strstr(cstr,"\\xmpd-win32.exe")) != NULL) {
	   *cptr = 0;
           if (*cstr == '\"')
	      cstr++;
           sprintf(xmpd_home_dir,"%s",cstr);
        }
        _chdir(xmpd_home_dir);
     }
#endif

     /* initialize authenUsers/profiles before reading config file */
     initUsers();     

     /* read config file (XML format!) (if applicable) */
     /* contents of config file overrule command-line options */

     /* config file should be read after command-line args? XXX */
     defaultManager[0] = 0;
     readXmpdConfigFile(xmpd_config_file);

     /* dumpUsers(); */

     /* now, process command-line args */
     /* process args 1..argc-1 */
     for (i=1; i<argc; i++) {
         if (argv[i][0] == '-') {

            switch (argv[i][1]) {
  	      case 'b':
		 xmpd_daemon = 1;
                 break;

 	      case 't':
		 xmpd_trace = 1;
                 break;

   	      case 'f':
		 if (i+1 < argc) {
		    xmpd_config_file = argv[++i];
                 }
                 break;

 	      case 'p':
		 if (i+1 < argc) {
		    xmpd_port = (u_short)atoi(argv[++i]);
	         }
                 break;
         
  	      case 'v':
		 xmpd_verbose = 1;
		 xmp_session_verbose = 1;
                 break;

              /* rest ignore because they were caught by main() */
   	      case 'h':
   	      case 's':
  	      default:
                 break;

            } /* switch */

         } /* if '-' */
     } /* for */

     /* daemonize ourselves if specified */
#ifndef WIN32
     if (xmpd_daemon != 0) {
        make_a_daemon_out_of_me();
     }
     else {
       /* not daemon but still need to handle some signals */
       mini_daemon();
     }
 
#endif

     /* now its safe to open log since we've determined where we live */
     open_log();
     xmpd_log("xmpd starting ------------");

     if (xmpd_verbose) {
        printf("xmpd: home dir is '%s'\n",xmpd_home_dir);
        sprintf(logBuf,"xmpd: home dir is '%s'",xmpd_home_dir);
        xmpd_log(logBuf);
     }
     sprintf(logBuf,"xmpd: home dir is '%s'",xmpd_home_dir);
     xmpd_log(logBuf);

     /* initialize our list of modules */
     init_modules();

     /* initialize ourselves -- we do all of this before 
        we start creating/starting threads */
     initVars();

     /* initialize core MIB objects prior to plugins */
     initCore();

     /* get our FQDN; coreMib() has figured it out and its in
        global variable sysName[] */

     /* load/initialize plugins and seed each with thread */
     initPlugins();
     findLoadAllPlugins(xmpd_home_dir);

     /* initalize SSL related parameters, keys, certs, etc. */
     /* we are assuming cwd[] is set */
   
     /* initialize before calling any thread stuff */

     if ((sslInit = init_ssl_stuff(sysName,"xmpd.pem","xmpd.csr")) < 0) {
        fprintf(stderr,"Failed to initialize SSL\n");
        sprintf(logBuf,"Failed to initialize SSL");
        xmpd_log(logBuf);
        exit(-1);
     }

     /* start connection-listener thread */
     if (pthread_create(&cl,NULL,xmpdListenForConnections,NULL) < 0) {
        perror("Failed to create connection-listener thread");
        sprintf(logBuf,"Failed to create connection-listener thread");
        xmpd_log(logBuf);
        exit(-1);
     }
     if (pthread_create(&hk,NULL,houseKeeping,NULL) < 0) {
        perror("Failed to create housekeeping thread");
        sprintf(logBuf,"Failed to create housekeeping thread");
        xmpd_log(logBuf);
        exit(-1);
     }

     if (xmpd_verbose) {
        printf("xmpd: finished initialization\n");
        sprintf(logBuf,"xmpd: finished initialization");
        xmpd_log(logBuf);
     }

     if (xmpd_verbose)
        xmpDumpVars();

     if (xmpd_verbose)
        xmpDumpVarsBackwards();

     /* send a coldstart trap */
#if 1
     sendColdstartTrap();
#endif

     /* join connection-listener and housekeeping threads */
     pthread_join(cl,NULL);
     pthread_join(hk,NULL);

     fprintf(stderr,"xmpd: joined threads, cleaning up\n");

#ifdef WIN32
     if (win32service) {
        UpdateStatus(-1,-1);
     }

     WSACleanup();

     if (win32service)
        UpdateStatus(-1,-1);
#endif

     if (xmpd_verbose) {
        printf("xmpd: finished\n");
        sprintf(logBuf,"xmpd: finished");
        xmpd_log(logBuf);
     }

     close_log();

     terminate_ssl_stuff();

     return 1;
}

/* ************************************************************* */

#ifdef WIN32

int uname(struct utsname *u)
{
     SYSTEM_INFO si;
     char nbname[128];
     char wsname[128];
     OSVERSIONINFOEX osi;
     int nbnamesz,IsWin64Platform();

     if (u == NULL)
        return -1;

     bzero(u,sizeof(struct utsname));
     bzero(&si,sizeof(si));
     GetSystemInfo(&si);

     /* system name (NetBIOS Name) */
     nbnamesz = 127;
     GetComputerName(nbname,&nbnamesz);
     gethostname(wsname,127);

     osi.dwOSVersionInfoSize = sizeof(osi);
     GetVersionEx(&osi);

#if 0
     printf("uname: major:%d minor:%d build:%d plat:%d sp:%d sp-:%d\n",
            osi.dwMajorVersion,
            osi.dwMinorVersion,
            osi.dwBuildNumber,
            osi.dwPlatformId,
            osi.wServicePackMajor,
            osi.wServicePackMinor);
#endif

     /* now, convert all of this numeric data into something
        a bit more human friendly */

     sprintf(u->sysname,"Windows",wsname);

     switch (osi.dwMajorVersion) {

       case 4: 
	  if (osi.dwMinorVersion == 0)
	     sprintf(u->version,"NT 4");
          else
	     sprintf(u->version,"NT 4.%d",osi.dwMinorVersion);
	  break;

       case 5:
	  switch (osi.dwMinorVersion) {
            case 0:
               sprintf(u->version,"Windows 2000");
	       break;
            case 1:
               sprintf(u->version,"Windows XP");
	       break;
            case 2:
               sprintf(u->version,"Windows Server 2003");
	       break;
            default:
               sprintf(u->version,"NT 5.%d",osi.dwMinorVersion);
	       break;
	  }
	  break;

       default:
	  sprintf(u->version,"NT %d.%d",osi.dwMajorVersion,osi.dwMinorVersion);
     }

     /* release */
     sprintf(u->release,"Build %d, SP%d.%d",
             osi.dwBuildNumber,
             osi.wServicePackMajor,
             osi.wServicePackMinor);

     /* nodename -- gethostname */
     sprintf(u->nodename,"%s",wsname);
     DBG1("Winsock hostname is '%s'\n",wsname);

     /* machine, hardware type from system info */
     sprintf(u->machine,"x86");
     switch (si.dwProcessorType) {
       case PROCESSOR_INTEL_386:
       case PROCESSOR_INTEL_486:
       case PROCESSOR_INTEL_PENTIUM:
       default:
 	  break;
     }

     if (IsWin64Platform()) {
        sprintf(u->machine,"x86_64");
     }

     return 0;
}

#endif /* WIN32 */

/* ************************************************************* */

int getApplicationName(char *appname, int maxsize)
{
     if (appname == NULL)
        return -1;
     appname[0] = 0;

#ifdef SOLARIS
#ifdef SOLX86
     sprintf(appname,"/xmpd-solx86");
#else
     sprintf(appname,"/xmpd-solaris");
#endif
#endif

#ifdef LINUX
     sprintf(appname,"/xmpd-linux");
#endif

#ifdef WIN32
     sprintf(appname,"\\xmpd-win32.exe");
#endif
     return 1;
}


/* ************************************************************* */
/* am I a win32 program running on a win64 system using WOW64? 
   In order to use IsWow64Process(), we need to be compiled for
   WinXP or higher (APPVER>=5.01).  We currently set APPVER=5.0
   for Win2k or higher.  So, we use GetProcAddress() to detect
   if IsWow64Process() is in kernel32.dll
*/

#ifdef WIN32

static FARPROC IsWow64Ptr = NULL;
static int IsWow64PtrInit = 0;

int IsWin64Platform()
{
     BOOL result;

     if ((IsWow64Ptr == NULL) && (IsWow64PtrInit == 0)) {

        IsWow64Ptr = GetProcAddress(GetModuleHandle("kernel32.dll"),
                                    "IsWow64Process");
        IsWow64PtrInit = 1;
     }

     /* if the function does not exist, we are not on a 64-bit platform 
        or even one capable of being 64-bit */
     if (IsWow64Ptr == NULL) 
        return 0;

     if ((IsWow64Ptr)(GetCurrentProcess(),&result) == FALSE) {
        return 0;
     }

     return result;
}

#endif

/* ************************************************************* */
